#version 410

out vec4 FragColor;

in vec3 worldPos;
in vec3 worldNormal;
in vec2 uv_MainTex;
in vec4 lightSpacePos;
in vec4 visibleSpacePos;

uniform vec4 m_color;
uniform float m_opacity;
uniform vec3 m_lightPos0; //world space, directional light
uniform vec3 m_lightColor0; 

uniform sampler2D _Diffuse_Map0;
uniform sampler2D _Shadow_Map0;
uniform sampler2D _Visible_Map0;

uniform bool is_Shadow_Map0;
uniform bool is_Diffuse_Map0;
uniform bool is_Visible_Map0;


float ShadowCalculation(vec4 fragPosLightSpace)
{
    // perform perspective divide
    vec3 projCoords = fragPosLightSpace.xyz / fragPosLightSpace.w;
    // transform to [0,1] range
    projCoords = projCoords * 0.5 + 0.5;
    // get closest depth value from light's perspective (using [0,1] range fragPosLight as coords)
    //float closestDepth = texture(_Shadow_Map0, projCoords.xy).r; 
    // get depth of current fragment from light's perspective
    float currentDepth = projCoords.z;
    // calculate bias (based on depth map resolution and slope)
    vec3 normal = normalize(worldNormal);
    vec3 lightDir = normalize(m_lightPos0);
    float bias = max(0.05 * (1.0 - dot(normal, lightDir)), 0.005);
    // check whether current frag pos is in shadow
    // float shadow = currentDepth - bias > closestDepth  ? 1.0 : 0.0;
    // PCF
    float shadow = 0.0;
    vec2 texelSize = 1.0 / textureSize(_Shadow_Map0, 0);
    for(int x = -1; x <= 1; ++x)
    {
        for(int y = -1; y <= 1; ++y)
        {
            float pcfDepth = texture(_Shadow_Map0, projCoords.xy + vec2(x, y) * texelSize).r; 
            shadow += currentDepth - bias > pcfDepth  ? 1.0 : 0.0;        
        }    
    }
    shadow /= 9.0;
    
    // keep the shadow at 0.0 when outside the far_plane region of the light's frustum.
    if(projCoords.z > 1.0)
        shadow = 0.0;
        
    return shadow;
}


float VisibleCalculation(vec4 fragPosMainCameraSpace)
{
    // perform perspective divide
    vec3 projCoords = fragPosMainCameraSpace.xyz / fragPosMainCameraSpace.w;
    // transform to [0,1] range
    projCoords = projCoords * 0.5 + 0.5;
    // get depth of current fragment from light's perspective
    float currentDepth = projCoords.z;
    // get closest depth value from light's perspective (using [0,1] range fragPosLight as coords)
    //float closestDepth = texture(_Visible_Map0, projCoords.xy).r; 
    float bias = 0.005;
    float visible = 0; // currentDepth - bias > closestDepth ? 0.0 : 1.0; 
    vec2 texelSize = 1.0 / textureSize(_Visible_Map0, 0);
    for(int x = -1; x <= 1; ++x)
    {
        for(int y = -1; y <= 1; ++y)
        {
            float pcfDepth = texture(_Visible_Map0, projCoords.xy + vec2(x, y) * texelSize).r; 
            visible += currentDepth - bias > pcfDepth  ? 0.0 : 1.0;        
        }    
    }
    visible /= 9.0;
    if(projCoords.z > 1.0)
        visible = 0.0;
    return visible;
}


void main()
{
    float visibility = 1;
    if(is_Visible_Map0){
        visibility = VisibleCalculation(visibleSpacePos);     
        if(visibility < 0.1){
            FragColor = m_color;
            return;
        }
    }

    vec4 textColor = is_Diffuse_Map0 ? texture(_Diffuse_Map0, uv_MainTex) : vec4(1, 1, 1, 1);
    float shadow = is_Shadow_Map0 ? ShadowCalculation(lightSpacePos) : 0;
    
    float ambientStrength = 0.6;  // ambient
    vec3 ambient = ambientStrength * m_lightColor0;

    float diffStrength = 0.7;
    vec3 norm = normalize(worldNormal);  // diffuse 
    vec3 lightDir = normalize(m_lightPos0); // normalize(m_lightPos - worldPos); 
    float diff = max(dot(norm, lightDir), 0.0);
    vec3 diffuse = diffStrength * diff * m_lightColor0;

    vec3 result = (ambient + (1-shadow) * diffuse) * textColor.rgb;
    result = result * visibility + m_color.rgb * (1-visibility);
    FragColor = clamp(vec4(result, m_opacity), 0.0, 1.0);
}
