#version 410

uniform vec4 m_color;  //main color, (0.6, 0.3 , 0.1, 1)
uniform float _Ka; //ambient strength, (0, 1), 0.1
uniform float _Kd; //diffuse strength, (0, 1), 0.1

uniform sampler2D _Alpha_Map0; //alpha texture
uniform sampler2D _Diffuse_Map0; //albedo texture
uniform sampler2D _AO_Map0; //AO texture
uniform sampler2D _Brightness_Map0; //brightness mask
uniform sampler2D _Random_Map0; //random tex
uniform sampler2D _Normal_Map0;  //normal map
uniform sampler2D _Shadow_Map0;


uniform bool is_Alpha_Map0;
uniform bool is_Diffuse_Map0;
uniform bool is_AO_Map0;
uniform bool is_Brightness_Map0;
uniform bool is_Random_Map0;
uniform bool is_Normal_Map0;
uniform bool is_Shadow_Map0;

uniform float _AO_Scale; //(0, 100), 5
uniform float _Cutoff; //(0, 1.02), 0.1
uniform float _Brightness; //brightness, (0.01,10), 1.0
uniform float _Randomness;  //(0-2), 1.0
uniform float _Bumpiness; //(0-5) 1.0

uniform float _R_Width; //(0, 0.3), 0.2
uniform float _R_Peak; //(0, 0.5), 0.1
uniform float _R_Cone_Angle; //Range(-10, 10), -5

uniform float _TT_Width; //(0, 0.5), 0.1
uniform float _TT_Peak; //(0, 1), 0.5
uniform float _TT_Cone_Angle; //(-10, 10), 2.5
uniform float _TT_Sharpness; //(1, 10), 5

uniform float _TRT_Width; //(0, 1), 0.4
uniform float _TRT_Peak; //(0, 1), 0.5
uniform float _TRT_Cone_Angle; //(-10, 10), 5
uniform float _Scatter_Amount; //(0, 2), 1
uniform float _Back_Scatter_Amount; //(0, 1000), 1
uniform float _Back_Scatter_Power; //(0.1, 200), 100

uniform vec3 m_viewPos;
// lightPos is currently in world space
uniform vec3 m_lightPos0; //directional light
uniform vec3 m_lightColor0; //directional light
uniform vec3 m_lightPos1; //point light
uniform vec3 m_lightColor1; //point light


in vec2 uv_MainTex;
in vec3 worldPos;
in vec3 worldNormal;
in vec3 worldTangent;
in vec4 lightSpacePos;
//in vec3 vColor;
in float lightWeight;

out vec4 FragColor;


//https://blog.selfshadow.com/publications/blending-in-detail/
vec3 RNM(vec3 base, vec3 detail)
{
    vec3 n1 = base.xyz;// *2 - 1;
    vec3 n2 = detail.xyz;// *2 - 1;

    vec3 b1;
    vec3 b2;
    vec3 b3 = n1;
    
    if (n1.z < -0.9999999) // Handle the singularity
    {
        b1 = vec3(0, -1, 0);
        b2 = vec3(-1, 0, 0);
    }
    else{
        float a = 1 / (1 + n1.z);
        float b = -n1.x*n1.y*a;
    
        // Form a basis
        b1 = vec3(1 - n1.x*n1.x*a, b, -n1.x);
        b2 = vec3(b, 1 - n1.y*n1.y*a, -n1.y);
    }
    
    // Rotate n2 via the basis
    vec3 r = n2.x*b1 + n2.y*b2 + n2.z*b3;
    
    return normalize(r);// r*0.5 + 0.5;
}


vec3 AmbientHair(float Ka, vec3 Cl) {
    return Ka * Cl;
}

vec3 KajiaKayHair(vec3 U, vec3 lightDir, float Kd, vec3 Cl) {
    // Wi: Light direction, pointing away from the object and towards the light
    vec3 Wi = normalize(lightDir);
    
    vec3 Wi_projected = normalize(Wi - dot(Wi, U) * U);
    float diffuse = clamp((dot(Wi, Wi_projected) + _Scatter_Amount) / (1 + _Scatter_Amount), 0, 1);
    
    return Kd * Cl * diffuse;
}


float
unit_height_gaussian(float variance, float dist_mean){
    return exp(-0.5 * dist_mean * dist_mean / variance);
}

float
M(float width_squared, float dist_mean) {
    return unit_height_gaussian(width_squared, dist_mean);
}

float
N_R(float phi) {
    return cos(phi/2.0);
}

float
N_TT(float phi, float sharpness)
{
    return unit_height_gaussian(1.0/(sharpness * sharpness), 3.14159265 - phi);
}

float
N_TRT(float phi) {
    // Ignoring the glints for simplicity.
    return cos(phi/2.0);
}

vec3 HairBRDF(vec3 U, vec3 lightDir, vec3 viewDir,
             vec3 R_Color, vec3 TT_Color, vec3 TRT_Color) {
    // -------------------------
    // Light independent
    // -------------------------
    
    vec3 Wi;     // Lighting direction
    vec3 Wr;     // Viewing direction
    
    vec3 Wi_projected;
    vec3 Wr_projected;
    
    float theta_i;        // incoming longitudinal angle
    float theta_r;        // outgoing longitudinal angle
    float theta_h;        // (theta_i + theta_r)/2
    float phi;            // azimuthal angle
    
    float cos_theta_i;
    float cos_theta_r;
    float cos_phi;
    
    float sign_theta_i;
    float sign_theta_r;
    
    // Wr: Viewing direction:
    // Pointing away from the object towards the eye
    Wr = normalize(viewDir);
    
    Wr_projected = normalize(Wr - dot(Wr, U) * U);
    
    cos_theta_r = dot(Wr, Wr_projected);
    sign_theta_r = sign(dot(Wr, U));
    theta_r = sign_theta_r * acos(cos_theta_r);
    
    // -------------------------
    // Light dependent
    // -------------------------
    
    // Wi: Light direction
    // Pointing away from the object and towards the light
    Wi = normalize(lightDir);
    
    Wi_projected = normalize(Wi - dot(Wi, U) * U);
    cos_theta_i = dot(Wi, Wi_projected);
    
    sign_theta_i = sign(dot(Wi, U));
    
    theta_i = sign_theta_i * acos(cos_theta_i);
    theta_h = (theta_i + theta_r) * 0.5;
    
    cos_phi = dot(Wr_projected, Wi_projected);
    phi = acos(cos_phi);
    
    vec3 result = vec3(0, 0, 0);
    
    // R
    result += R_Color.rgb * _R_Peak *
    M(_R_Width * _R_Width, theta_h - radians(_R_Cone_Angle)) *
    N_R(phi);
    // TT
    result += TT_Color.rgb * _TT_Peak *
    M(_TT_Width * _TT_Width, theta_h - radians(_TT_Cone_Angle)) *
    N_TT(phi, _TT_Sharpness);
    // TRT
    result += TRT_Color.rgb * _TRT_Peak *
    M(_TRT_Width * _TRT_Width, theta_h - radians(_TRT_Cone_Angle)) *
    N_TRT(phi);
    
    return result;
}


vec3 HairShaderRGB
(
 vec3 normalDir,
 vec3 tangentDir,
 vec3 lightDir,
 vec3 viewDir,
 float shadow,
 vec3 Albedo,
 float Ka, // Ambient
 vec3 Ambient_Color,
 float Kd, // Diffuse
 vec3 Diffuse_Color,
 // Hair: TT, TRT
 vec3 R_Color,
 vec3 TT_Color,
 vec3 TRT_Color,
 vec2 uv,
 vec4 rand4
 )
{
    // Hair tangent toward the tip computed from the surface normal and the surface tangent.
    // The negation is due to Unity's left handed system.
    //vec3 U = -normalize(cross(normalDir, tangentDir));
    vec3 U = normalize(cross(normalDir, tangentDir));
    vec3 V = tangentDir;
    
    // Adding variation to hair direction
    U += rand4.rgb * U;
    U = normalize(U);
    // R_Color *= AmbientHair(Ka, Ambient_Color);
    // TT_Color *= fixed3(0, 1, 0);
    // TRT_Color *= fixed3(0, 0, 1);
    
    vec3 amibient_scattering = AmbientHair(Ka, Ambient_Color);
    vec3 diffuse_scattering = KajiaKayHair(U, lightDir, Kd, Diffuse_Color);
    vec3 single_scattering = HairBRDF(U, lightDir, viewDir, R_Color, TT_Color, TRT_Color);
    
    vec3 total_scattering = amibient_scattering + (diffuse_scattering + single_scattering) * (1-shadow);
    
    vec3 result = total_scattering * Albedo;
    return result;
}



float ShadowCalculation(vec4 fragPosLightSpace)
{
    // perform perspective divide
    vec3 projCoords = fragPosLightSpace.xyz / fragPosLightSpace.w;
    // transform to [0,1] range
    projCoords = projCoords * 0.5 + 0.5;
    // get closest depth value from light's perspective (using [0,1] range fragPosLight as coords)
    // float closestDepth = texture(_Shadow_Map0, projCoords.xy).r; 
    // get depth of current fragment from light's perspective
    float currentDepth = projCoords.z;
    // calculate bias (based on depth map resolution and slope)
    vec3 normal = normalize(worldNormal);
    vec3 lightDir = m_lightPos0;
    float bias = max(0.05 * (1.0 - dot(normal, lightDir)), 0.005);
    // check whether current frag pos is in shadow
    // float shadow = currentDepth - bias > closestDepth  ? 1.0 : 0.0;
    // PCF
    float shadow = 0.0;
    vec2 texelSize = 1.0 / textureSize(_Shadow_Map0, 0);
    for(int x = -1; x <= 1; ++x)
    {
        for(int y = -1; y <= 1; ++y)
        {
            float pcfDepth = texture(_Shadow_Map0, projCoords.xy + vec2(x, y) * texelSize).r; 
            shadow += currentDepth - bias > pcfDepth  ? 1.0 : 0.0;        
        }    
    }
    shadow /= 9.0;
    
    // keep the shadow at 0.0 when outside the far_plane region of the light's frustum.
    if(projCoords.z > 1.0)
        shadow = 0.0;
        
    return shadow;
}


void main()
{
    float Alpha = is_Alpha_Map0 ? texture(_Alpha_Map0, uv_MainTex).r : 1;
    if(Alpha < _Cutoff)
        discard;

    vec3 _Color = lightWeight * m_color.rgb;  //vColor; // * (1-m_color.a) + m_color.rgb * m_color.a;

    vec3 Albedo = vec3(1,1,1); // albedo_tex.rgb;
    if(is_Diffuse_Map0)
        Albedo = texture(_Diffuse_Map0, uv_MainTex).rgb;

    vec3 brightness = vec3(1,1,1);
    if(is_Brightness_Map0)
        brightness = texture(_Brightness_Map0, uv_MainTex).rgb;

    float AO = _AO_Scale;
    if(is_AO_Map0)
        AO = texture(_AO_Map0, uv_MainTex).r * _AO_Scale;

    vec4 rand4 = vec4(0, 0, 0, 1);
    if(is_Random_Map0)
        rand4 = texture(_Random_Map0, uv_MainTex) * _Randomness;

    vec3 Normal = normalize(worldNormal);
    if(is_Normal_Map0){
        vec3 detailNormal = texture(_Normal_Map0, uv_MainTex).xyz * 2 - 1;
        detailNormal.rg *= _Bumpiness;
        Normal = RNM(worldNormal, detailNormal);
    }
    
    vec3 Tangent = normalize(worldTangent);


    float shadow = 0;
    if(is_Shadow_Map0){
        shadow = ShadowCalculation(lightSpacePos);
    }

    //half tangentSign = o.surface_tangent.w * unity_WorldTransformParams.w;
    //vec3 biTangent = normalize(cross(Normal, surface_tangent));
    //surface_tangent.xyz = normalize(cross(biTangent, Normal));
    
    vec3 viewDir = normalize(m_viewPos - worldPos);

    //directional light
    vec3 lightDir0 = normalize(m_lightPos0); //directional light
    vec3 lightColor0 = m_lightColor0 * _Brightness;
    vec3 result_rgb0 = HairShaderRGB
    (
     Normal,
     Tangent,
     lightDir0,
     viewDir,
     shadow, //shadow
     Albedo,
     // Ambient
     _Ka * AO,
     _Color * lightColor0,
     // Diffuse
     _Kd,
     _Color * lightColor0,
     // Hair: R, TT, TRT
     lightColor0,
     _Color * lightColor0,
     _Color * lightColor0,
     uv_MainTex,
     rand4
     );
    
    vec3 result = result_rgb0 * brightness;
    FragColor = vec4(clamp(result,0.0,1.0), m_color.a);
    //FragColor = vec4(clamp(result,0.0,1.0), Alpha);
}



